/*
 * Ratiometric Frequency Counter
 *
 * displays 6 digits from 1Hz up to ~40MHz
 * cycle time 4M/F_CPU ... 16M/F_CPU
 * theoretical lower frequency limit: F_CPU/16M
 * theoretical upper frequency limit: 4*F_CPU
 *
 * (c) Axel (XL) Schwenke, axel.schwenke@gmx.net
 *
 * $Id: main.c 113 2010-12-10 14:23:02Z schwenke $
 */

#define VERSION "V1.2 (c)XL"
#define DEBUG 0
#define SLEEP 1     /* lets save energy ;) */
#define WDT   1     /* enable hardware watchdog */

#if F_CPU > (16*1024*1024-512)
#error "F_CPU is too big! Must fit in 24 bits"
#endif

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <util/delay.h>

#include "arith.h"
#include "lcd.h"

/* prototypes */

void hw_init(void);
void hexout(uint8_t);

/* macros */

#if DEBUG
#define phase1()    lcd_gotoxy(0,0)
#define phase2()    lcd_gotoxy(1,0)
#define phase3()    lcd_gotoxy(2,0)
#define underflow() lcd_clrscr(); lcd_puts("underflow!"); _delay_ms(2000)
#else
#define phase1()
#define phase2()
#define phase3()
#define underflow() lcd_gotoxy(4,1); lcd_putc('Z')
#endif

/* global variables */

volatile union {    /* 24 bit counter f_x */
    uint8_t  b[3];
} nx;

volatile union {    /* 24 bit counter f_ref */
    uint16_t w[1];
    uint8_t  b[3];
} nref;

volatile union {    /* 48 bit result */
    uint32_t q[1];
    uint8_t  b[6];
} res;

EEMEM uint32_t fref = F_CPU;

/* the code */

int main(void)
{
    hw_init();
#if DEBUG
    lcd_init(LCD_DISP_ON_CURSOR);
#else
    lcd_init(LCD_DISP_ON);
#endif

    lcd_puts(VERSION);
    _delay_ms(1000);

    /* display current reference freq */
    res.q[0]= eeprom_read_dword(&fref);
    mybin2bcd((void*) &res);
    lcd_clrscr();
    lcd_putc('R');
    lcd_putc('=');
    hexout(res.b[3]);
    hexout(res.b[2]);
    hexout(res.b[1]);
    hexout(res.b[0]);
    _delay_ms(2000);

    /* display current adjustment */
    res.q[0]= ADC;
    lcd_clrscr();
    lcd_puts("ADJ=\n");
    if (res.q[0] == 0) {
        lcd_puts("(off)");
    } else {
        if (res.q[0] >= 512) {
            lcd_putc('+');
            res.q[0] -= 512;
        } else {
            lcd_putc('-');
            res.q[0] = 512 - res.q[0];
        }
        mybin2bcd((void*) &res);
        hexout(res.b[1]);
        hexout(res.b[0]);
    }
    _delay_ms(2000);

#if WDT && !DEBUG
    wdt_enable(WDTO_2S);
#endif

    while (1) {

        /* clear counters */
        clear_low();
        nx.b[2]= 0;
        TCNT0= 0;
        clear_high();

        phase1(); /* wait for gate to open */

        TCCR1B &= ~_BV(ICES1);  /* capture falling edge */
        TIMSK  |=  _BV(TOIE1);  /* enable Timer1 overflow irq */
        sei();
        start_high();
        while (gate_is_closed()) {
            wdt_reset();
            sleep_cpu();
        }

        phase2(); /* wait for ~4M fref pulses */

        while (nref.b[2]<65) {
            wdt_reset();
            sleep_cpu();
        }

        phase3(); /* wait for gate to close */

        TCCR1B |=  _BV(ICES1);  /* capture rising edge */
        start_low();
        while (nref.b[2] && gate_is_open()) {
            wdt_reset();
            sleep_cpu();
        }

        cli();

        if (nref.b[2] == 0) {
            /* f_ref counter overflowed */
            underflow();

        } else {
            uint8_t prefix=' '; /* unit prefix ' 'Hz */
            uint8_t decimal=7;  /* decimal point after digit #7 */

            /* read fx counter */
            lat_oe_low();
            lat_oe_high();      /* rising edge latches 74HC590 */
            lat_oe_low();       /* low enables 74HC590 drivers */

            register union {    /* ugly workaround */
                uint16_t w;     /* to avoid even uglier gcc code */
                uint8_t  b[2];
            } tmp;

            tmp.b[0]= TCNT0;
            tmp.b[1]= nx.b[2];
            tmp.w    <<= 1;
            tmp.b[0] |= fx9_read();

            nx.b[0] = data_read();
            nx.b[1] = tmp.b[0];
            nx.b[2] = tmp.b[1];
            lat_oe_high();

#if DEBUG
            lcd_clrscr();
            lcd_putc('x');
            lcd_putc('=');
            hexout(nx.b[2]);
            hexout(nx.b[1]);
            hexout(nx.b[0]);
            _delay_ms(2000);

            lcd_clrscr();
            lcd_putc('r');
            lcd_putc('=');
            hexout(nref.b[2]);
            hexout(nref.b[1]);
            hexout(nref.b[0]);
            _delay_ms(2000);
#endif

            /* res = fref + calibration */
            res.q[0]  = eeprom_read_dword(&fref);
            register uint32_t cal= ADC;
            if (cal != 0) {
                res.q[0] += (cal - 512);
            }

            /* res *= nx */
            mymul((void*) &res, (void*) &nx);

            /* multiply by 10 until big enough
             * this shifts the decimal position */
            while ((res.b[5] & 0xF0) == 0 && decimal > 1) {
                /* res *= 10 */
                mymulby10((void*) &res);
                --decimal;
            }

            /* res /= nref */
            mydiv((void*) &res, (void*) &nref);

            /* res = PackedBCD(res) */
            mybin2bcd((void*) &res);

            /* skip leading zeros */
            while (decimal && (res.b[3] & 0xF0) == 0) {
                res.q[0] <<= 4;
                --decimal;
            }

            /* adjust unit prefix */
            if (decimal>2) {
                prefix   = 'k';
                decimal -= 3;
            }
            if (decimal>2) {
                prefix   = 'M';
                decimal -= 3;
            }

            /* write 6 digits + decimal separator */
            lcd_clrscr();
            for (register uint8_t i=0; i<6; i++) {
                lcd_putc('0' + ((res.b[3] & 0xF0) >> 4));
                res.q[0] <<= 4;
                if (i==decimal) {
                    lcd_putc('.');
                }
            }

            /* append unit */
            lcd_putc(prefix);
            lcd_putc('H');
            lcd_putc('z');

#if DEBUG
            _delay_ms(3000);
#endif
        }
    }
}


void hw_init(void)
{
    /* pullups enabled on all inputs */
    /* data bus pins are inputs */
    /* START=L, CLR=L, LAT/OE=H */

    PORTB= ~(_BV(CLEAR_PIN) | _BV(START_PIN));
    DDRB=   (_BV(CLEAR_PIN) | _BV(START_PIN));

    PORTC= ~(_BV(ANALOG_PIN));
    DDRC=  0;

    PORTD= ~(_BV(LCD_E_PIN));
    DDRD=   (_BV(LCD_RS_PIN) | _BV(LCD_RW_PIN) | _BV(LCD_E_PIN) | _BV(LAT_OE_PIN));

    /* disable Analog Comparator */

    ACSR = (_BV(ACD));

    /* ADC senses PC5 continuously, Vref=Aref */

    ADMUX=  (_BV(MUX2) | _BV(MUX0));
    ADCSRA= (_BV(ADEN) | _BV(ADSC) | _BV(ADFR) | _BV(ADPS2) | _BV(ADPS1) | _BV(ADPS0));

    /* Timer 0: ext. trigger, falling edge */
    /* Timer 1: system clock */

    TCCR0=  (_BV(CS02) | _BV(CS01));
    TCCR1A= 0;
    TCCR1B= (_BV(CS10));

    /* Timer 0: capture irq and
     * Timer 1: overflow irq enabled */

    TIMSK=  (_BV(TICIE1) | _BV(TOIE0));

#if SLEEP
    set_sleep_mode(SLEEP_MODE_IDLE);
    sleep_enable();
#endif

}


void hexout(uint8_t x)
{
    register uint8_t d= (x>>4) & 0x0F;
    lcd_putc(d + (d>9 ? 'A'-10 : '0'));
    d= x & 0x0F;
    lcd_putc(d + (d>9 ? 'A'-10 : '0'));
}


/* Interrupt Service Routines */

/* Timer 0 Overflow - f_x counter */
ISR(TIMER0_OVF_vect)
{
    ++nx.b[2];
}

/* Timer 1 Overflow - f_ref counter */
ISR(TIMER1_OVF_vect)
{
    ++nref.b[2];
}

/* Timer 1 Capture - gate open/close */
ISR(TIMER1_CAPT_vect)
{
    if ((TCCR1B & _BV(ICES1)) == 0) {
        /* falling edge = gate opens */
        nref.w[0]= ICR1;
        nref.b[2]= 0;
    } else {
        /* rising edge = gate closes */
        register uint16_t tmp= ICR1;
        if ((TIFR & _BV(TOV1)) && tmp < 256) {
            /* overflow irq pending - possible collision */
            ++nref.b[2];
        }
        if (tmp < nref.w[0]) {
            --nref.b[2];
        }
        nref.w[0]= tmp - nref.w[0];
        TIMSK &= ~_BV(TOIE1);  /* disable and */
        TIFR   =  _BV(TOV1);   /* clear overflow interrupt */
    }
}

